# -*- coding: utf-8 -*-
from cvnn.utils import *
import numpy as np
import sys
from math import sqrt
from scipy import signal
from pdb import set_trace
from abc import ABC, abstractmethod
import plotly.graph_objects as go
import plotly
from matplotlib import pyplot as plt
from scipy.linalg import eigh, cholesky
from scipy.stats import norm
import tikzplotlib

MARKERS = [".", "x", "s", "+", "^", "D", "_", "v", "|", "*", "H"]
logger = logging.getLogger(cvnn.__name__)


# =======
# Dataset
# =======


class Dataset:
    """
    This class is used to centralize all dataset management.
    """

    def __init__(self, x, y, num_classes=None, ratio=0.8, savedata=False, batch_size=None,
                 categorical=True, debug=False, shuffle=False, dataset_name=""):
        """
        :param x: Data
        :param y: Labels/outputs
        :param num_classes: Number of different classes to be made
        :param ratio: (float) [0, 1]. Percentage of Tran case vs Test case (ratio = #train / (#train + #test))
            Default: 0.8 (80% of the data will be used for train).
        :param savedata: (boolean) If true it will save the generated data into "./data/current/date/path/".
            Default: False
        """
        self.dataset_name = dataset_name
        self.random_shuffle = shuffle
        x = np.array(x)
        y = np.array(y)
        if x.dtype == np.complex128:  # Do this cast not to have warning messages when fit
            x = x.astype(np.complex64)
        elif x.dtype == np.float64:
            x = x.astype(np.float32)
        self.x = x
        self.y = y.astype(np.float32)
        self.categorical = categorical  # TODO: know it automatically as done in other functions
        if categorical:
            self.y = self.sparse_into_categorical(self.y)
        if num_classes is None:
            self.num_classes = self._deduce_num_classes()  # This is only used for plotting the data example
        else:
            self.num_classes = num_classes
        self.ratio = ratio
        self.save_path = "./data/"
        # Generate data from x and y
        self.x_test, self.y_test = None, None  # Tests
        self.x_train, self.y_train = None, None  # Train
        self._generate_data_from_base()
        if savedata:
            self.save_data()
        # Parameters used with the fit method
        self._iteration = 0
        if batch_size is None:
            self.batch_size = self.x_train.shape[0]  # Don't use batches at all
        else:
            if np.shape(self.x_train)[0] < batch_size:  # TODO: make this case work as well. Just display a warning
                logger.error("Batch size was bigger than total amount of examples")
                sys.exit(-1)
            self.batch_size = batch_size
        if debug:
            self.plot_data(overlapped=True, showfig=True, save_path=None)

    def get_next_batch(self):
        num_tr_iter = int(self.x_train.shape[0] / self.batch_size)  # Number of training iterations in each epoch
        if not self._iteration < num_tr_iter:
            logger.error("I did more calls to this function that planned")
            sys.exit(-1)
        # Get the next batch
        start = self._iteration * self.batch_size
        end = (self._iteration + 1) * self.batch_size
        self._iteration += 1

        return self._get_next_batch(self.x_train, self.y_train, start, end)

    def _deduce_num_classes(self):
        """
        Tries to deduce the total amount of classes present in the network.
        ATTENTION: This method will obviously fail for regression data.
        """
        # https://jovianlin.io/cat-crossentropy-vs-sparse-cat-crossentropy/
        if len(self.y.shape) == 1:  # Sparse labels
            num_samples = max(self.y.astype(int)) - min(self.y.astype(int)) + 1
        else:  # Categorical labels
            num_samples = self.y.shape[1]
        return num_samples

    def _generate_data_from_base(self):
        """
        Generates everything (x_test, y_test, x_train, y_train, x_real) once x and y is defined.
        """
        self.x_train, self.y_train, self.x_test, self.y_test = self.separate_into_train_and_test(self.x, self.y,
                                                                                                 self.ratio,
                                                                                                 pre_rand=self.random_shuffle)

    def shuffle(self):
        """
        Shuffles the train data and reset iteration counter
        """
        self.x_train, self.y_train = randomize(self.x_train, self.y_train)
        self._iteration = 0

    def save_data(self, save_path=None):
        """
        Saves data into the specified path as a numpy array.
        """
        if save_path is None:
            save_path = create_folder(self.save_path)
        else:
            os.makedirs(save_path, exist_ok=True)
            save_path = Path(save_path)
        if os.path.exists(save_path):
            np.save(save_path / "data.npy", self.x)
            np.save(save_path / "labels.npy", self.y)
            # Save also an image of the example
            self.plot_data(overlapped=True, showfig=False, save_path=save_path)
        else:
            logger.error("Path {} does not exist".format(save_path))

    def summary(self, res_str=None):
        """
        :return: String with the information of the dataset.
        """
        if res_str is None:
            res_str = self.dataset_name
        res_str += "\tNum classes: {}\n".format(self.num_classes)
        res_str += "\tTotal Samples: {}\n".format(self.x.shape[0])
        res_str += "\tVector size: {}\n".format(self.x.shape[1])
        res_str += "\tTrain percentage: {}%\n".format(int(self.ratio * 100))
        return res_str

    def plot_data(self, overlapped=False, showfig=True, save_path=None, library='matplotlib'):
        """
        Generates a figure with an example of the data
        :param overlapped: (boolean) If True it will plot all the examples in the same figure changing the color.
            Otherwise it will create a subplot with as many subplots as classes. Default: False
        :return: (fig, ax) matplolib format to be plotted.
        """
        if library == 'matplotlib':
            self._plot_data_matplotlib(overlapped=overlapped, showfig=showfig, save_path=save_path)
        elif library == 'plotly':
            self._plot_data_plotly(showfig=showfig, save_path=save_path)
        else:
            logger.warning("Unrecognized library to plot " + library)
            return None

    def _plot_data_plotly(self, showfig=False, save_path=None, extension=".svg"):
        fig = go.Figure()
        labels = self.y
        if self.categorical:
            labels = self.categorical_to_sparse(labels)
        for cls in range(self.num_classes):
            for index, label in enumerate(labels):
                if label == cls:
                    fig.add_trace(go.Scatter(x=np.real(self.x[index]), y=np.imag(self.x[index]),
                                             mode='markers', marker_symbol=cls, marker_size=10,
                                             name="Class: " + str(cls)))
                    break
        fig.update_layout(title='Data Visualization Example',
                          yaxis=dict(scaleanchor="x", scaleratio=1),
                          xaxis_title='real (x)',
                          yaxis_title='imaginary (y)',
                          showlegend=True)
        if save_path is not None:
            # https://plot.ly/python/configuration-options/
            os.makedirs(save_path, exist_ok=True)
            plotly.offline.plot(fig, filename=str(save_path / "data_example.html"), config={'editable': True},
                                auto_open=showfig)
            fig.write_image(str(save_path / "data_example") + extension)
        elif showfig:
            fig.show(config={'scrollZoom': True, 'editable': True})

    def _plot_data_matplotlib(self, overlapped=False, showfig=False, save_path=None, extension=".svg"):
        labels = self.y
        if self.categorical:
            labels = self.categorical_to_sparse(labels)
        if overlapped:
            fig, ax = plt.subplots()
            for cls in range(self.num_classes):
                for index, label in enumerate(labels):
                    if label == cls:
                        ax.plot(np.real(self.x[index]),
                                np.imag(self.x[index]), MARKERS[cls % len(MARKERS)], label="Class " + str(cls))
                        ax.axis('equal')
                        ax.grid(True)
                        ax.set_aspect('equal', adjustable='box')
                        break
            ax.set_xlabel('real (x)')
            ax.set_ylabel('imaginary (y)')
            ax.legend(loc='upper right')
        else:
            fig, ax = plt.subplots(self.num_classes)
            for cls in range(self.num_classes):
                for index, label in enumerate(labels):
                    if label == cls:  # This is done in case the data is shuffled.
                        ax[cls].plot(np.real(self.x[index]),
                                     np.imag(self.x[index]), 'b.')
                        ax[cls].axis('equal')
                        ax[cls].grid(True)
                        ax[cls].set_aspect('equal', adjustable='box')
                        break
        if showfig:
            fig.show()
        if save_path is not None:
            os.makedirs(save_path, exist_ok=True)
            prefix = ""
            if overlapped:
                prefix = "overlapped_"
            fig.savefig(save_path / Path(prefix + "data_example" + extension), transparent=True)
            tikzplotlib.save(save_path / (prefix + "data_example.tikz"))
        return fig, ax

    # =======
    # Getters
    # =======

    def get_train_and_test(self):
        return self.x_train, self.y_train, self.x_test, self.y_test

    def get_test(self):
        return self.x_test, self.y_test

    def get_all(self):
        return self.x, self.y

    def get_categorical_labels(self):
        return self.sparse_into_categorical(self.y, self.num_classes)

    # ================
    # Static functions
    # ================

    @staticmethod
    def sparse_into_categorical(spar, num_classes=None):
        if len(spar.shape) == 1:  # Check data is indeed sparse
            spar = spar.astype(int)
            if num_classes is None:
                num_classes = max(spar) + 1  # assumes labels starts at 0
            cat = np.zeros((spar.shape[0], num_classes))
            for i, k in enumerate(spar):
                cat[i][k] = 1
        else:
            # Data was already categorical (I think)
            cat = spar
        return cat

    @staticmethod
    def categorical_to_sparse(cat):
        return np.argmax(cat, axis=1)

    @staticmethod
    def separate_into_train_and_test(x, y, ratio=0.8, pre_rand=True):
        """
        Separates data x with corresponding labels y into train and test set.
        :param x: data
        :param y: labels of data x
        :param ratio: value between 0 and 1.
            1 meaning all the data x will be the training set and 0 meaning all data x will be the test set.
        :param pre_rand: if True then x and y will be shuffled first (maintaining coherence between them)
        :return: tuple (x_train, y_train, x_test, y_test) of the training and test set both data and labels.
        """
        if (ratio > 1) or (ratio < 0):
            sys.exit("Error:separate_into_train_and_test: ratio should be between 0 and 1. Got value " + str(ratio))
        if pre_rand:
            x, y = randomize(x, y)
        m = np.shape(x)[0]
        x_train = x[:int(m * ratio)]
        y_train = y[:int(m * ratio)]
        x_test = x[int(m * ratio):]
        y_test = y[int(m * ratio):]
        return x_train, y_train, x_test, y_test

    @staticmethod
    def _get_next_batch(x, y, start, end):
        """
        Get next batch from x and y using start and end
        :param x: data
        :param y: data labels
        :param start: starting index of the batch to be returned
        :param end: end index of the batch to be returned (not including)
        :return: tuple (x, y) of the selected batch
        """
        if start < 0:
            sys.exit("Error:get_next_batch(): start parameter cannot be negative")
        if start > end:  # What will happen if not? Should I leave this case anyway and just give a warning?
            sys.exit("Error:get_next_batch(): end should be higher than start")
        # TODO: Check end < len(x)
        x_batch = x[start:end]
        y_batch = y[start:end]
        return x_batch, y_batch


class OpenDataset(Dataset):
    """
    This class is used to init a Dataset with a saved npy data instead of giving the vector directly.
    Construction overload (either use vector x and y or use a string path)
        will be maybe cleaner but it does not exist in Python :S
    """

    def __init__(self, path, num_classes=None, ratio=0.8, savedata=False):
        self.path = cast_to_path(path)
        x, y = self.load_dataset(self.path)
        super().__init__(x, y, num_classes=num_classes, ratio=ratio, savedata=savedata,
                         dataset_name="opened dataset " + str(self.path))

    @staticmethod
    def load_dataset(path):
        try:
            x = np.load(path / "data.npy")
            y = np.load(path / "labels.npy")
        except FileNotFoundError:
            sys.exit("OpenDataset::load_dataset: Files data.npy and labels.npy not found in " + path)
        return x, y

    def summary(self, res_str=None):
        res_str = "Opened data located in {}\n".format(str(self.path))
        return super().summary(res_str)


class GeneratorDataset(ABC, Dataset):
    """
    Is a database method with an automatic x and y (data) generation.
    Used to automate the generation of data.
    Must therefore define a method to generate the data.

    Good Practice: Although it is not compulsory,
        it is recommended to define it's own summary method to know how the dataset was generated.
    """

    def __init__(self, m, n, num_classes=2, ratio=0.8, savedata=False, debug=False, dataset_name=None):
        """
        This class will first generate x and y with it's own defined method and then initialize a conventional dataset
        """
        x, y = self._generate_data(m, n, num_classes)
        x, y = randomize(x, y)
        if dataset_name is None:
            dataset_name = "Generated dataset"
        Dataset.__init__(self, x, y, num_classes=num_classes, ratio=ratio, savedata=savedata, debug=debug,
                         dataset_name=dataset_name)

    @abstractmethod
    def _generate_data(self, num_samples_per_class, num_samples, num_classes):
        """
        Abstract method. It MUST be defined.
        Method on how to generate x and y.
        """
        pass


class CorrelatedGaussianNormal(GeneratorDataset):

    def __init__(self, m, n, cov_matrix_list, num_classes=None, ratio=0.8, debug=False, savedata=False):
        if num_classes is None:
            num_classes = len(cov_matrix_list)
        if not len(cov_matrix_list) == num_classes:
            logger.error("cov_matrix_list length ({0}) should have the same size as num_classes ({1})".format(
                len(cov_matrix_list), num_classes))
            sys.exit(-1)
        for cov_mat in cov_matrix_list:  # Each class has a coviariance matrix 2x2
            # Numpy cast enables data to be either numpy array or list
            if not np.array(cov_mat).shape == (2, 2):
                logger.error("covariance matrix must have shape 2x2 but has shape {}".format(np.array(cov_mat).shape))
                sys.exit(-1)
            if not cov_mat[0][1] == cov_mat[1][0]:
                logger.error("Elements outside the diagonal must be equal (they are both sigma_{xy}")
                sys.exit(-1)
            if not np.abs(cov_mat[0][1] / sqrt(cov_mat[0][0] * cov_mat[1][1])) < 1:
                logger.error("corelation coefficient module must be lower than one")
                sys.exit(-1)
        self.cov_matrix_list = cov_matrix_list
        super().__init__(m, n, num_classes=num_classes, ratio=ratio, savedata=savedata, debug=debug,
                         dataset_name="Correlated Gaussian Normal")

    @staticmethod
    def _create_correlated_gaussian_point(num_samples, r=None):
        # https: // scipy - cookbook.readthedocs.io / items / CorrelatedRandomSamples.html
        # Choice of cholesky or eigenvector method.
        method = 'cholesky'
        # method = 'eigenvectors'
        if r is None:
            # The desired covariance matrix.
            r = np.array([
                [1, 1.41],
                [1.41, 2]
            ])
        # Generate samples from three independent normally distributed random
        # variables (with mean 0 and std. dev. 1).
        x = norm.rvs(size=(2, num_samples))

        # We need a matrix `c` for which `c*c^T = r`.  We can use, for example,
        # the Cholesky decomposition, or the we can construct `c` from the
        # eigenvectors and eigenvalues.
        if method == 'cholesky':
            # Compute the Cholesky decomposition.
            c = cholesky(r, lower=True)
        else:
            # Compute the eigenvalues and eigenvectors.
            evals, evecs = eigh(r)
            # Construct c, so c*c^T = r.
            c = np.dot(evecs, np.diag(np.sqrt(evals)))

        # Convert the data to correlated random variables.
        y = np.dot(c, x)
        return [y[0][i] + 1j * y[1][i] for i in range(y.shape[1])]

    def _generate_data(self, num_samples_per_class, num_samples, num_classes):
        x = []
        y = []
        for signal_class in range(num_classes):
            r = self.cov_matrix_list[signal_class]
            y.extend(signal_class * np.ones(num_samples_per_class))
            for _ in range(num_samples_per_class):
                x.append(self._create_correlated_gaussian_point(num_samples, r))
        return np.array(x), np.array(y)

    def summary(self, res_str=None):
        res_str = "Correlated Gaussian Noise\n"
        for cls in range(self.num_classes):
            res_str += "class {}\n".format(cls)
            res_str += "\tPearson correlation coefficient: {}\n".format(self.get_coef_correl(cls))
            res_str += "\tsigma_x^2 = " + str(self.cov_matrix_list[cls][0][0]) + \
                       "sigma_y^2 = " + str(self.cov_matrix_list[cls][1][1])
            res_str += "\tCircularity quotient: {}\n".format(self.get_circularity_quotient(cls))
            variance, pseudo_variance = self.get_variance_and_pseudo_variance(cls)
            res_str += "\t\tvariance: {0}; pseudo-variance: {1}\n".format(variance, pseudo_variance)
            epsilon, alpha = self.get_ellipse_params(cls, deg=True)
            res_str += "\tEllipse epsilon: {0}, Angle (alpha): {1} deg\n".format(epsilon, alpha)
        return super().summary(res_str)

    # =====================================================
    # Get all the parameter equivalents
    # https://ieeexplore.ieee.org/abstract/document/4682548
    # =====================================================

    def get_coef_correl(self, index):
        rho = None
        if index < len(self.cov_matrix_list):
            cov_mat = self.cov_matrix_list[index]
            rho = cov_mat[0][1] / (sqrt(cov_mat[0][0]) * sqrt(cov_mat[1][1]))
        else:
            logger.error("Index out of range")
        return rho

    def get_variance_and_pseudo_variance(self, index):
        variance = None
        pseudo_variance = None
        if index < len(self.cov_matrix_list):  # TODO: Get the error here
            variance = self.cov_matrix_list[index][0][0] + self.cov_matrix_list[index][1][1]
            pseudo_variance = self.cov_matrix_list[index][0][0] - self.cov_matrix_list[index][1][1] + \
                              2j * self.cov_matrix_list[index][0][1]
        return variance, pseudo_variance

    def get_circularity_quotient(self, index):
        variance, pseudo_variance = self.get_variance_and_pseudo_variance(index)
        return pseudo_variance / variance

    def get_ellipse_params(self, index, deg=False):
        varrho = self.get_circularity_quotient(index)
        epsilon = sqrt(np.abs(varrho))
        alpha = np.angle(varrho, deg)
        return epsilon, alpha


class CorrelatedGaussianCoeffCorrel(CorrelatedGaussianNormal):

    def __init__(self, m, n, param_list, num_classes=None, ratio=0.8, debug=False, savedata=False):
        if num_classes is None:
            num_classes = len(param_list)
        if not len(param_list) == num_classes:
            logger.error("param_list length ({0}) should have the same "
                         "size as num_classes ({1})".format(len(param_list), num_classes))
        cov_mat_list = []
        for param in param_list:
            if not len(param) == 3:
                logger.error("Each parameter in param_list should have size 3 "
                             "(coef correl and both variances) but {} where given".format(len(param)))
                sys.exit(-1)
            sigma_xy = param[0] * sqrt(param[1] * param[2])
            cov_mat_list.append([[param[1], sigma_xy], [sigma_xy, param[2]]])
        super().__init__(m=m, n=n, cov_matrix_list=cov_mat_list,
                         num_classes=num_classes, ratio=ratio, debug=debug, savedata=savedata)


class ComplexNormalVariable(CorrelatedGaussianNormal):
    """
    This class is a correlated gaussian normal but instead of defining it's classes with the covariance matrix
    it is defined with it's complex variance and pseudo-variance.
    https://ieeexplore.ieee.org/abstract/document/4682548
    """

    def __init__(self, m, n, param_list, num_classes=None, ratio=0.8, debug=False, savedata=False):
        if num_classes is None:
            num_classes = len(param_list)
        if not len(param_list) == num_classes:
            logger.error("param_list length ({0}) should have the same size "
                         "as num_classes ({1})".format(len(param_list), num_classes))
            sys.exit(-1)
        cov_mat_list = []
        for param in param_list:
            if not len(param) == 2:
                logger.error("Each parameter in param_list should have size 2 "
                             "(sigma and tau) but {} where given".format(len(param)))
                sys.exit(-1)
            cov_mat_list.append(self.get_cov_matrix(param[0], param[1]))
        super().__init__(m=m, n=n, cov_matrix_list=cov_mat_list,
                         num_classes=num_classes, ratio=ratio, debug=debug, savedata=savedata)
        for i, param in enumerate(param_list):  # Just for fun
            assert self.get_circularity_quiotient(i) == param[1] / param[0], \
                "ComplexNormalVariable::__init__: Error in creating data"

    @staticmethod
    def get_cov_matrix(sigma, tau):
        sigma_xx = (sigma + np.real(tau)) / 2
        sigma_yy = (sigma - np.real(tau)) / 2
        sigma_xy = np.imag(tau) / 2
        return [[sigma_xx, sigma_xy], [sigma_xy, sigma_yy]]


class GaussianNoise(GeneratorDataset):

    def __init__(self, m, n, num_classes=2, ratio=0.8, savedata=False, function='hilbert'):
        noise_gen_dispatcher = {
            'non_correlated': self._create_non_correlated_gaussian_noise,
            'hilbert': self._create_hilbert_gaussian_noise
        }
        try:
            self.function = noise_gen_dispatcher[function]
        except KeyError:
            sys.exit("GaussianNoise: Unknown type of noise" + str(function))
        super().__init__(m, n, num_classes=num_classes, ratio=ratio, savedata=savedata, dataset_name="Gaussian Noise")

    def _generate_data(self, num_samples_per_class, num_samples, num_classes):
        x = np.empty((num_classes * num_samples_per_class, num_samples)) \
            + 1j * np.empty((num_classes * num_samples_per_class, num_samples))
        # I am using zeros instead of empty because although counter intuitive it seams it works faster:
        # https://stackoverflow.com/questions/55145592/performance-of-np-empty-np-zeros-and-np-ones
        # DEBUNKED? https://stackoverflow.com/questions/52262147/speed-of-np-empty-vs-np-zeros?
        # Initialize all at 0 to later put a 1 on the corresponding place
        # TODO: generate to zero the other parameters as well
        y = np.zeros((num_classes * num_samples_per_class, num_classes))

        for k in range(num_classes):
            mu = int(100 * np.random.rand())
            sigma = 15 * np.random.rand()
            logger.info("Class " + str(k) + ": mu = " + str(mu) + "; sigma = " + str(sigma))
            x[k * num_samples_per_class:(k + 1) * num_samples_per_class, :] = self.function(num_samples_per_class,
                                                                                            num_samples, mu, sigma)
            y[k * num_samples_per_class:(k + 1) * num_samples_per_class, k] = 1
        return normalize(x), y

    @staticmethod
    def _create_non_correlated_gaussian_noise(num_samples_per_class, num_samples, mu, sigma):
        """
        Creates a numpy matrix of size mxn with random gaussian distribution of mean mu and variance sigma
        """
        return (np.random.normal(mu, sigma, (num_samples_per_class, num_samples)) +
                1j * np.random.normal(mu, sigma, (num_samples_per_class, num_samples))) / sqrt(2)

    @staticmethod
    def _create_hilbert_gaussian_noise(num_samples_per_class, num_samples, mu, sigma):
        x_real = np.random.normal(mu, sigma, (num_samples_per_class, num_samples))
        return signal.hilbert(x_real)

    def summary(self, res_str=None):
        res_str = "Gaussian {} Noise\n".format(str(self.function).replace('_', ' '))
        return super().summary(res_str)


# =================
# Testing Functions
# =================


def create_subplots_of_graph():
    # monte_carlo_loss_gaussian_noise(iterations=100, filename="historgram_gaussian.csv")
    m = 5
    n = 100
    num_classes = 2
    coefs = [0.1, 0.4, 0.75, 0.999]
    overlaped = True
    rows = 2
    if overlaped:
        rows = 1
    fig, axs = plt.subplots(rows, len(coefs), sharex=True, sharey=True)
    # , gridspec_kw={'hspace': 0, 'wspace': 0})
    for i, coef in enumerate(coefs):
        dataset = CorrelatedGaussianNormal(m, n, num_classes=num_classes, debug=False, cov_matrix_list=coef)
        x, y = dataset.get_all()
        for r in range(rows):
            for cls in range(num_classes):
                for index, label in enumerate(y):
                    if label == cls:
                        if overlaped:
                            axs[i].plot(np.real(x[index]), np.imag(x[index]), MARKERS[cls % len(MARKERS)])
                            axs[i].axis('equal')
                            axs[i].grid(True)
                            axs[i].set_aspect('equal', adjustable='box')
                            break
                        else:
                            axs[r, i].plot(np.real(x[index]), np.imag(x[index]), 'b.')
                            axs[r, i].axis('equal')
                            axs[r, i].grid(True)
                            axs[r, i].set_aspect('equal', adjustable='box')
                            break
    if overlaped:
        for ax, coef in zip(axs, coefs):
            ax.set_title("coef abs: {}".format(coef))
        for cls, ax in enumerate(axs):
            ax.set_ylabel("class {}".format(int(cls)), size='large')
    else:
        for ax, coef in zip(axs[0], coefs):
            ax.set_title("coef abs: {}".format(coef))
        for cls, ax in enumerate(axs[:, 0]):
            ax.set_ylabel("class {}".format(int(cls)), size='large')
    fig.show()
    # create_correlated_gaussian_noise(n, debug=True)
    # set_trace()


def sup(a, b):
    return a > b


def inf(a, b):
    return a > b


def parametric_predictor(dataset, coef_1=0.5, coef_2=-0.5):
    rho = []
    x = np.real(dataset.x)
    y = np.imag(dataset.x)
    for re, im in zip(x, y):
        cov = np.cov([re, im])
        rho.append(cov[0][1] / (cov[0][0] * cov[1][1]))

    rho = np.array(rho)
    thresh = np.full(rho.shape, (coef_1 + coef_2) / 2)
    if coef_1 > coef_2:
        result = np.less(rho, thresh).astype(int)
    else:
        result = np.greater(rho, thresh).astype(int)

    acc = np.sum(np.equal(Dataset.categorical_to_sparse(dataset.y), result)) / len(result)
    return acc


if __name__ == "__main__":
    # create_subplots_of_graph()
    m = 10000
    n = 128
    """cov_matr_list = [
        [[1, 0.75], [0.75, 1]],
        [[1, -0.75], [-0.75, 1]]
    ]
    dataset = CorrelatedGaussianNormal(m, n, cov_matr_list, debug=False)"""
    dataset = CorrelatedGaussianCoeffCorrel(m, n, param_list=[[0.5, 1, 1], [-0.5, 1, 1]])
    dataset.save_data("./data/MLSP/")

    # dataset = OpenDataset("./data/MLSP/")
    # dataset.plot_data(overlapped=True, showfig=True, library="matplotlib")
    # set_trace()
    # print("{:.2%}".format(parametric_predictor(dataset)))

__author__ = 'J. Agustin BARRACHINA'
__version__ = '0.1.20'
__maintainer__ = 'J. Agustin BARRACHINA'
__email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'



from setuptools import setup
import versioneer

requirements = [
    'tensorflow>=2.0',
    'numpy',
    'pandas', 'scipy', 'prettytable',                   # Data
    'colorlog', 'openpyxl',                             # Logging
    'matplotlib', 'seaborn', 'plotly', 'tikzplotlib',   # Plotting
    'tqdm'
]

setup(
    name='cvnn',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Library to help implement a complex-valued neural network (cvnn) using tensorflow as back-end",
    license="MIT",
    author="J Agustin BARRACHINA",
    author_email='joseagustin.barra@gmail.com',
    url='https://github.com/NEGU93/cvnn',
    packages=['cvnn'],
    entry_points={
        'console_scripts': [
            'cvnn=cvnn.cli:cli'
        ]
    },
    install_requires=requirements,
    keywords='cvnn',
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ]
)



import tensorflow as tf

"""
# Loss functions
# https://ml-cheatsheet.readthedocs.io/en/latest/loss_functions.html

All loss functions will be declared here. 
Loss functions will be type agnostic, meaning it can be with either complex or real value outputs and labels.
"""

"""---------
# REGRESSION
---------"""


def mean_square(y, y_out):
    """
    Mean Squared Error, or L2 loss.
    :return:
    """
    with tf.compat.v1.name_scope("loss"):
        error = y - y_out
        return tf.square(tf.abs(error))


"""-------------
# CLASSIFICATION
-------------"""


def categorical_crossentropy(y, y_out):
    """
    https://jovianlin.io/cat-crossentropy-vs-sparse-cat-crossentropy/
    https://stats.stackexchange.com/questions/260505/should-i-use-a-categorical-cross-entropy-or-binary-cross-entropy-loss-for-binary
    :return: -y*log(y_out)-(1-y)*log(1-y_out) where:
        log - the natural log
        y - binary indicator (0 or 1), it will be all 0's but one (according to the corresponding class)
        y_out - predicted probability observation the class
    """
    # TODO: not working for the moment
    with tf.compat.v1.name_scope("loss_scope") as scope:
        y1_error = tf.math.multiply(-y, tf.math.log(y_out))  # Error for y = 1
        y0_error = tf.math.multiply(1 - y, tf.math.log(1 - y_out))  # Error for y = 0
        error = tf.keras.backend.sum(tf.math.subtract(y1_error, y0_error), 1)   # Sum of every
        return error


__author__ = 'J. Agustin BARRACHINA'
__copyright__ = 'Copyright 2020, {project_name}'
__credits__ = ['{credit_list}']
__license__ = '{license}'
__version__ = '0.1.1'
__maintainer__ = 'J. Agustin BARRACHINA'
__email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'
__status__ = '{dev_status}'



import cvnn.dataset as dp
from utils import randomize
from cvnn.layers import Dense
from cvnn.cvnn_model import CvnnModel
import tensorflow as tf
from utils import transform_to_real
import numpy as np
from pdb import set_trace

if __name__ == "__main__":
    m = 10000
    n = 100
    num_classes = 2
    dataset = dp.CorrelatedGaussianNormal(m, n, num_classes, debug=True, cov_matrix_list=0.75)
    x, y = dataset.get_all()
    x = x.astype(np.complex64)
    y = dp.Dataset.sparse_into_categorical(y, num_classes=num_classes).astype(np.float32)
    x_train, y_train, x_test, y_test = dp.Dataset.separate_into_train_and_test(x, y)
    x_train_real = transform_to_real(x_train)
    x_test_real = transform_to_real(x_test)

    epochs = 100
    batch_size = 100
    display_freq = 160
    learning_rate = 0.002

    # Create complex network
    input_size = x.shape[1]  # Size of input
    output_size = y.shape[1]  # Size of output
    h1_size = 100
    h2_size = 40
    shape = [Dense(input_size=input_size, output_size=h1_size, activation='cart_relu',
                   input_dtype=np.complex64, output_dtype=np.complex64),
             Dense(input_size=h1_size, output_size=h2_size, activation='cart_relu',
                   input_dtype=np.complex64, output_dtype=np.complex64),
             Dense(input_size=h2_size, output_size=output_size, activation='cart_softmax_real',
                   input_dtype=np.complex64, output_dtype=np.float32)]
    complex_network = CvnnModel(name="complex_network", shape=shape, loss_fun=tf.keras.losses.categorical_crossentropy,
                                verbose=True, tensorboard=True, save_csv_checkpoints=True)
    complex_network.fit(x_train, y_train, x_test, y_test, learning_rate=learning_rate, epochs=epochs,
                        batch_size=batch_size, display_freq=display_freq)
    complex_network.plotter.plot_key(key='accuracy', library='plotly', name=complex_network.name)
    # Create real network   TODO: interesting to do a cast option like the deepcopy right?
    input_size *= 2
    h1_size *= 2
    h2_size *= 2
    shape = [Dense(input_size=input_size, output_size=h1_size, activation='cart_relu',
                   input_dtype=np.float32, output_dtype=np.float32),
             Dense(input_size=h1_size, output_size=h2_size, activation='cart_relu',
                   input_dtype=np.float32, output_dtype=np.float32),
             Dense(input_size=h2_size, output_size=output_size, activation='cart_softmax_real',
                   input_dtype=np.float32, output_dtype=np.float32)]
    real_network = CvnnModel(name="real_network", shape=shape, loss_fun=tf.keras.losses.categorical_crossentropy,
                             verbose=True, tensorboard=True, save_csv_checkpoints=True)

    # Train both networks
    real_network.fit(x_train_real, y_train, x_test_real, y_test, learning_rate=learning_rate, epochs=epochs,
                     batch_size=batch_size, display_freq=display_freq)
    real_network.plotter.plot_key(key='accuracy', library='plotly', name=real_network.name)

__author__ = 'J. Agustin BARRACHINA'
__version__ = '0.0.20'
__maintainer__ = 'J. Agustin BARRACHINA'
__email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'



from abc import ABC, abstractmethod
from itertools import count
import tensorflow as tf
from collections.abc import Iterable
import sys
import cvnn
import logging
from typing import Union, Optional, Callable          # https://docs.python.org/3/library/typing.html
from cvnn.activation_functions import apply_activation
from cvnn.utils import get_func_name
import numpy as np
from time import time
from pdb import set_trace
import cvnn.initializers as initializers

SUPPORTED_DTYPES = (np.complex64, np.float32)  # , np.complex128, np.float64) Gradients return None when complex128
layer_count = count(0)  # Used to count the number of layers

t_Dtype = Union[tf.dtypes.DType, np.dtype]
t_Shape = Union[int, tuple, list]
t_Callable_shape = Union[t_Shape, Callable]


class ComplexLayer(ABC):
    # Being ComplexLayer an abstract class, then this can be called using:
    #   self.__class__.__bases__.<variable>
    # As all child's will have this class as base, mro gives a full list so won't work.
    last_layer_output_dtype = None      # TODO: Make it work both with np and tf dtypes
    last_layer_output_size = None

    def __init__(self, output_size: t_Callable_shape, input_size=Optional[t_Shape], input_dtype=t_Dtype, **args):
        """
        Base constructor for a complex layer. The first layer will need a input_dtype and input_size.
        For the other classes is optional,
            if input_size or input_dtype does not match last layer it will throw a warning
        :param output_size: Output size of the layer.
            If the output size depends on the input_size, a function must be passed as output_size.
        :param input_size: Input size of the layer
        :param input_dtype: data type of the input
        """
        self.logger = logging.getLogger(cvnn.__name__)
        if output_size is None:
            self.logger.error("Output size = None not supported")
            sys.exit(-1)

        if input_dtype is None and self.__class__.__bases__[0].last_layer_output_dtype is None:
            # None input dtype given but it's the first layer declared
            self.logger.error("First layer must be given an input dtype", exc_info=True)
            sys.exit(-1)
        elif input_dtype is None and self.__class__.__bases__[0].last_layer_output_dtype is not None:
            # Use automatic mode
            self.input_dtype = self.__class__.__bases__[0].last_layer_output_dtype
        elif input_dtype is not None:
            if input_dtype not in SUPPORTED_DTYPES:
                self.logger.error("Layer::__init__: unsupported input_dtype " + str(input_dtype), exc_info=True)
                sys.exit(-1)
            if self.__class__.__bases__[0].last_layer_output_dtype is not None:
                if self.__class__.__bases__[0].last_layer_output_dtype != input_dtype:
                    self.logger.warning("Input dtype " + str(input_dtype) +
                                        " is not equal to last layer's input dtype " +
                                        str(self.__class__.__bases__[0].last_layer_output_dtype))
            self.input_dtype = input_dtype

        # This will be normally the case.
        # Each layer must change this value if needed.
        self.__class__.__bases__[0].last_layer_output_dtype = self.input_dtype

        # Input Size
        if input_size is None:
            if self.__class__.__bases__[0].last_layer_output_size is None:
                # None input size given but it's the first layer declared
                self.logger.error("First layer must be given an input size")
                sys.exit(-1)
            else:  # self.__class__.__bases__[0].last_layer_output_dtype is not None:
                self.input_size = self.__class__.__bases__[0].last_layer_output_size
        elif input_size is not None:
            if self.__class__.__bases__[0].last_layer_output_size is not None:
                if input_size != self.__class__.__bases__[0].last_layer_output_size:
                    self.logger.warning("Input size " + str(input_size) + " is not equal to last layer's output size " +
                                        str(self.__class__.__bases__[0].last_layer_output_size))
            self.input_size = input_size

        if callable(output_size):
            output_size()
            assert self.output_size is not None, "Error: output_size function must set self.output_size"
        else:
            self.output_size = output_size
        for x in self.__class__.mro():
            if x == ComplexLayer:
                x.last_layer_output_size = self.output_size
        # self.__class__.__bases__[0].last_layer_output_size = self.output_size
        self.layer_number = next(layer_count)  # Know it's own number
        self.__class__.__call__ = self.call     # Make my object callable

    @abstractmethod
    def __deepcopy__(self, memodict=None):
        pass

    def get_input_dtype(self):
        return self.input_dtype

    @abstractmethod
    def get_real_equivalent(self):
        """
        :return: Gets a real-valued COPY of the Complex Layer.
        """
        pass

    @abstractmethod
    def get_description(self) -> str:
        """
        :return: a string containing all the information of the layer
        """
        pass

    def _save_tensorboard_output(self, x, summary, step):
        x = self.call(x)
        with summary.as_default():
            if x.dtype == tf.complex64 or x.dtype == tf.complex128:
                tf.summary.histogram(name="Activation_value_" + str(self.layer_number) + "_real",
                                     data=tf.math.real(x), step=step)
                tf.summary.histogram(name="Activation_value_" + str(self.layer_number) + "_imag",
                                     data=tf.math.imag(x), step=step)
            elif x.dtype == tf.float32 or x.dtype == tf.float64:
                tf.summary.histogram(name="Activation_value_" + str(self.layer_number),
                                     data=x, step=step)
            else:
                self.logger.error("Input_dtype not supported. Should never have gotten here!", exc_info=True)
                sys.exit(-1)
        return x

    def save_tensorboard_checkpoint(self, x, weight_summary, activation_summary, step=None):
        self._save_tensorboard_weight(weight_summary, step)
        return self._save_tensorboard_output(x, activation_summary, step)

    @abstractmethod
    def _save_tensorboard_weight(self, weight_summary, step):
        pass

    @abstractmethod
    def trainable_variables(self):
        pass

    @abstractmethod
    def call(self, inputs):
        pass

    def get_output_shape_description(self) -> str:
        # output_string = ""
        if isinstance(self.output_size, Iterable):
            output_string = "(None, " + ", ".join([str(x) for x in self.output_size]) + ")"
        else:
            output_string = "(None, " + str(self.output_size) + ")"
        return output_string


class Flatten(ComplexLayer):

    def __init__(self, input_size=None, input_dtype=None):
        # Win x2: giving None as input_size will also make sure Flatten is not the first layer
        super().__init__(input_size=input_size, output_size=self._get_output_size, input_dtype=input_dtype)

    def __deepcopy__(self, memodict=None):
        return Flatten()

    def _get_output_size(self):
        self.output_size = np.prod(self.input_size)

    def get_real_equivalent(self):
        return self.__deepcopy__()

    def get_description(self) -> str:
        return "Complex Flatten"

    def _save_tensorboard_weight(self, weight_summary, step):
        return None

    def call(self, inputs):
        return tf.reshape(inputs, (inputs.shape[0], self.output_size))

    def trainable_variables(self):
        return []


class Dense(ComplexLayer):
    """
    Fully connected complex-valued layer
    Implements the operation:
        activation(dot(input, weights) + bias)
    - where data types can be either complex or real.
    - activation is the element-wise activation function passed as the activation argument,
    - weights is a matrix created by the layer
    - bias is a bias vector created by the layer
    """

    def __init__(self, output_size, input_size=None, activation=None, input_dtype=None,
                 weight_initializer=None, bias_initializer=None,
                 dropout=None):
        """
        Initializer of the Dense layer
        :param output_size: Output size of the layer
        :param input_size: Input size of the layer
        :param activation: Activation function to be used.
            Can be either the function from cvnn.activation or tensorflow.python.keras.activations
            or a string as listed in act_dispatcher
        :param input_dtype: data type of the input. Default: np.complex64
            Supported data types:
                - np.complex64
                - np.float32
        :param weight_initializer: Initializer for the weights.
            Default: tensorflow.keras.initializers.GlorotUniform
        :param bias_initializer: Initializer fot the bias.
            Default: tensorflow.keras.initializers.Zeros
        :param dropout: Either None (default) and no dropout will be applied or a scalar
            that will be the probability that each element is dropped.
            Example: setting rate=0.1 would drop 10% of input elements.
        """
        super(Dense, self).__init__(output_size=output_size, input_size=input_size, input_dtype=input_dtype)
        if activation is None:
            activation = 'linear'
        self.activation = activation
        # Test if the activation function changes datatype or not...
        self.__class__.__bases__[0].last_layer_output_dtype = \
            apply_activation(self.activation,
                             tf.cast(tf.complex([[1., 1.], [1., 1.]], [[1., 1.], [1., 1.]]), self.input_dtype)
                             ).numpy().dtype
        self.dropout = dropout
        if weight_initializer is None:
            weight_initializer = initializers.GlorotUniform()
        self.weight_initializer = weight_initializer
        if bias_initializer is None:
            bias_initializer = initializers.Zeros()
        self.bias_initializer = bias_initializer
        self.w = None
        self.b = None
        self._init_weights()

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        return Dense(output_size=self.output_size, input_size=self.input_size,
                     activation=self.activation,
                     input_dtype=self.input_dtype,
                     weight_initializer=self.weight_initializer,
                     bias_initializer=self.bias_initializer, dropout=self.dropout
                     )

    def get_real_equivalent(self, output_multiplier=2, input_multiplier = 2):
        """
        :param output_multiplier: Multiplier of output and input size (normally by 2)
        :return: real-valued copy of self
        """
        return Dense(output_size=int(round(self.output_size * output_multiplier)),
                     input_size=int(round(self.input_size * input_multiplier)),
                     activation=self.activation, input_dtype=np.float32,
                     weight_initializer=self.weight_initializer,
                     bias_initializer=self.bias_initializer, dropout=self.dropout
                     )

    def _init_weights(self):
        self.w = tf.Variable(self.weight_initializer(shape=(self.input_size, self.output_size), dtype=self.input_dtype),
                             name="weights" + str(self.layer_number))
        self.b = tf.Variable(self.bias_initializer(shape=self.output_size, dtype=self.input_dtype),
                             name="bias" + str(self.layer_number))

    def get_description(self):
        fun_name = get_func_name(self.activation)
        out_str = "Dense layer:\n\tinput size = " + str(self.input_size) + "(" + str(self.input_dtype) + \
                  ") -> output size = " + str(self.output_size) + \
                  ";\n\tact_fun = " + fun_name + ";\n\tweight init = " \
                  + self.weight_initializer.__name__ + "; bias init = " + self.bias_initializer.__name__ + \
                  "\n\tDropout: " + str(self.dropout) + "\n"
        return out_str

    def call(self, inputs):
        """
        Applies the layer to an input
        :param inputs: input
        :param kwargs:
        :return: result of applying the layer to the inputs
        """
        # TODO: treat bias as a weight. It might optimize training (no add operation, only mult)
        with tf.name_scope("ComplexDense_" + str(self.layer_number)) as scope:
            if tf.dtypes.as_dtype(inputs.dtype) is not tf.dtypes.as_dtype(np.dtype(self.input_dtype)):
                self.logger.warning("Dense::apply_layer: Input dtype " + str(inputs.dtype) + " is not as expected ("
                                    + str(tf.dtypes.as_dtype(np.dtype(self.input_dtype))) +
                                    "). Casting input but you most likely have a bug")
            out = tf.add(tf.matmul(tf.cast(inputs, self.input_dtype), self.w), self.b)
            y_out = apply_activation(self.activation, out)

            if self.dropout:
                # $ tf.nn.dropout(tf.complex(x,x), rate=0.5)
                # *** ValueError: x has to be a floating point tensor since it's going to be scaled.
                # Got a <dtype: 'complex64'> tensor instead.
                drop_filter = tf.nn.dropout(tf.ones(y_out.shape), rate=self.dropout)
                y_out_real = tf.multiply(drop_filter, tf.math.real(y_out))
                y_out_imag = tf.multiply(drop_filter, tf.math.imag(y_out))
                y_out = tf.cast(tf.complex(y_out_real, y_out_imag), dtype=y_out.dtype)
            return y_out

    def _save_tensorboard_weight(self, summary, step):
        with summary.as_default():
            if self.input_dtype == np.complex64 or self.input_dtype == np.complex128:
                tf.summary.histogram(name="ComplexDense_" + str(self.layer_number) + "_w_real",
                                     data=tf.math.real(self.w), step=step)
                tf.summary.histogram(name="ComplexDense_" + str(self.layer_number) + "_w_imag",
                                     data=tf.math.imag(self.w), step=step)
                tf.summary.histogram(name="ComplexDense_" + str(self.layer_number) + "_b_real",
                                     data=tf.math.real(self.b), step=step)
                tf.summary.histogram(name="ComplexDense_" + str(self.layer_number) + "_b_imag",
                                     data=tf.math.imag(self.b), step=step)
            elif self.input_dtype == np.float32 or self.input_dtype == np.float64:
                tf.summary.histogram(name="ComplexDense_" + str(self.layer_number) + "_w",
                                     data=self.w, step=step)
                tf.summary.histogram(name="ComplexDense_" + str(self.layer_number) + "_b",
                                     data=self.b, step=step)
            else:
                # This case should never happen. The constructor should already have checked this
                self.logger.error("Input_dtype not supported.", exc_info=True)
                sys.exit(-1)

    def trainable_variables(self):
        return [self.w, self.b]


class Dropout(ComplexLayer):

    def __init__(self, rate, noise_shape=None, seed=None):
        """
        :param rate: A scalar Tensor with the same type as x.
            The probability that each element is dropped.
            For example, setting rate=0.1 would drop 10% of input elements.
        :param noise_shape: A 1-D Tensor of type int32, representing the shape for randomly generated keep/drop flags.
        :param seed:  A Python integer. Used to create random seeds. See tf.random.set_seed for behavior.
        """
        # tf.random.set_seed(seed)
        self.rate = rate
        self.noise_shape = noise_shape
        self.seed = seed
        # Win x2: giving None as input_size will also make sure Dropout is not the first layer
        super().__init__(input_size=None, output_size=self.dummy, input_dtype=None)

    def dummy(self):
        self.output_size = self.input_size

    def call(self, inputs):
        drop_filter = tf.nn.dropout(tf.ones(inputs.shape), rate=self.rate, noise_shape=self.noise_shape, seed=self.seed)
        y_out_real = tf.multiply(drop_filter, tf.math.real(inputs))
        y_out_imag = tf.multiply(drop_filter, tf.math.imag(inputs))
        return tf.cast(tf.complex(y_out_real, y_out_imag), dtype=inputs.dtype)

    def _save_tensorboard_weight(self, weight_summary, step):
        # No tensorboard things to save
        return None

    def get_description(self):
        return "Complex Dropout:\n\trate={}".format(self.rate)

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        return Dropout(rate=self.rate, noise_shape=self.noise_shape, seed=self.seed)

    def get_real_equivalent(self):
        return self.__deepcopy__()      # Dropout layer is dtype agnostic

    def trainable_variables(self):
        return []


if __name__ == "__main__":
    import pdb; pdb.set_trace()


__author__ = 'J. Agustin BARRACHINA'
__copyright__ = 'Copyright 2020, {project_name}'
__credits__ = ['{credit_list}']
__license__ = '{license}'
__version__ = '0.0.25'
__maintainer__ = 'J. Agustin BARRACHINA'
__email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'
__status__ = '{dev_status}'
